# -*- coding: utf-8 -*-
"""
@author: Philippe@loco-labs.io

The `test_ntv_numpy` module contains the unit tests (class unittest) for the
`Darray`, `Ndarray` and `Xndarray` classes.
"""
import unittest
import requests

from decimal import Decimal
import numpy as np
from datetime import date, time
import pandas as pd
from shapely.geometry import Point, LineString
import xarray as xr

import ntv_pandas as npd

#from ntv_numpy.numpy_ntv_connector import read_json_tab, to_json_tab
from ntv_numpy import NdarrayConnec, XndarrayConnec
from ntv_numpy import Darray, Dfull, Dcomplete, Ndarray, Xndarray, Nutil, Xdataset, Dutil
 
from json_ntv import NtvConnector, Ntv, NtvSingle, NtvList
SeriesConnec = NtvConnector.connector()['SeriesConnec']
DataFrameConnec = NtvConnector.connector()['DataFrameConnec']
nd_equals = Dutil.equals

#read_json({':ndarray': ['int64', [1, 2]]})

class Test_Darray(unittest.TestCase):
    
    def test_darray_simple(self):
        
        example =[
            ([1, 2], 'Dfull'),
            ([[1, 2], [0, 1]], 'Dcomplete'),
            ([[10, 20], [1, 2]], 'Dfull'),
            ([[[10, 20], [1, 2]], [0, 1]], 'Dcomplete')
        ]

        for index, ex in enumerate(example):
            da = Darray.read_json(ex[0])
            self.assertEqual(da.__class__.__name__, ex[1])            
            self.assertEqual(len(da), len(ex[0]))            
            match ex[1]:
                case 'Dfull':
                    self.assertIsNone(da.ref)
                    self.assertTrue(nd_equals(np.array(None), da.coding))
                    self.assertTrue(nd_equals(da.data, da.values))           
                case 'Dcomplete':
                    da_full = Darray.read_json(example[index-1][0])
                    self.assertIsNone(da.ref)
                    self.assertFalse(nd_equals(np.array(None), da.coding))
                    self.assertTrue(nd_equals(da_full.values, da.values))   
        

    def test_darray_dtype(self):
        
        self.assertEqual(Darray.read_json([1, 'two'], dtype='object').to_json(),
                         [1, 'two'])
        
    def test_darray_nested(self):

        example =[
            np.array([np.array([1, 2], dtype='int64'), 
                       np.array(['test1', 'test2'], dtype='str_')],
                      dtype='object')
        ]

        for ex in example:
            da = Dfull(ex)
            #print(da)
            self.assertEqual(len(da), len(ex))            
            self.assertIsNone(da.ref)
            self.assertTrue(nd_equals(np.array(None), da.coding))
            self.assertTrue(nd_equals(da.data, da.values))           
    
class Test_Ndarray(unittest.TestCase):    

    def test_update(self):
        nda = Ndarray.read_json([[2,2],[1,2,3,4]])
        nda1 = Ndarray.read_json([[2,2],[1,2,3,4]])
        self.assertTrue(nda.update(Ndarray.read_json(['ex_nda2'])))
        self.assertTrue(nda.update(Ndarray.read_json([[2,2],[1,2,3,4]])))
        self.assertEqual(nda, nda1) 
        
    def test_set_array_uri(self):
        ndas = Ndarray.read_json([[2, 2], 'uri'])
        self.assertFalse(ndas.set_array([1, 2, 3]))
        self.assertTrue(ndas.set_array([[1, 2], [4, 3]]))
        self.assertFalse(ndas.set_array([[1.1, 2], [4, 3]]))
        self.assertTrue(ndas.set_array([[10, 20], [40, 30]]))
        self.assertFalse(ndas.set_array([[10, 20], [40, 30], [40, 30]]))
        self.assertTrue(ndas.set_uri('uri', no_ntv_type=True))
        self.assertEqual(ndas, Ndarray.read_json([[2, 2], 'uri']))
        
    def test_ndarray_null(self):    
        
        example =[[[], None],
                  ]

        for ex in example:
            #print(ex[0], ex[1])
            arr = Ndarray(ex[0], ntv_type=ex[1])
            js = arr.to_json()
            #print(js)
            ex_rt = Ndarray.read_json(js)
            self.assertEqual(ex_rt, arr)        
            self.assertEqual(js, ex_rt.to_json())         
            ex_rt = Ndarray.read_json(js, convert=False)
            #print(js, ex_rt.to_json(format=format))
            self.assertEqual(js[':ndarray'][1], ex_rt.to_json()[':ndarray'][1])         
            #print(np.array_equal(ex_rt, arr),  ex_rt, ex_rt.dtype)

    def test_ndarray_simple2(self):    
        
        example =[[[1,2], 'int64'],
                  [[1,2], None],
                  [[True, False], 'boolean'],
                  #[['1+2j', 1], 'complex'],
                  [['test1', 'test2'], 'string'], 
                  [['2022-01-01T10:05:21.0002', '2023-01-01T10:05:21.0002'], 'datetime'],
                  [['2022-01-01', '2023-01-01'], 'date'],
                  [['2022-01', '2023-01'], 'yearmonth'],
                  [['2022', '2023'], 'year'],
                  #[[1,2], 'timedelta[D]'],
                  [[b'abc\x09', b'abc'], 'base16'],
                  [[time(10, 2, 3), time(20, 2, 3)], 'time'],
                  [[{'one':1}, {'two':2}], 'object'],
                  [[None, None], 'null'],
                  [[Decimal('10.5'), Decimal('20.5')], 'decimal64'],
                  [[Point([1,2]), Point([3,4])], 'point'],
                  #[[Ntv.obj({':point':[1,2]}), NtvSingle(12, 'noon', 'hour')], 'ntv'],
                  [[LineString([[0, 0], [0, 1], [1, 1], [0, 0]]), 
                    LineString([[0, 0], [0, 10], [10, 10], [0, 0]])], 'line']
                  ]

        for ex in example:
            #print(ex[0], ex[1])
            arr = Ndarray(ex[0], ntv_type=ex[1])
            for format in ['full', 'complete']:
                js = arr.to_json(format=format)
                #print(js)
                ex_rt = Ndarray.read_json(js)
                self.assertTrue(ex_rt.shape == arr.shape == [2])        
                self.assertEqual(ex_rt, arr)        
                self.assertEqual(js, ex_rt.to_json(format=format))         
                ex_rt = Ndarray.read_json(js, convert=False)
                #print(js, ex_rt.to_json(format=format))
                self.assertEqual(js[':ndarray'][1], ex_rt.to_json(format=format)[':ndarray'][1])         
                #print(np.array_equal(ex_rt, arr),  ex_rt, ex_rt.dtype)
            if len(ex[0]) == 2:
                arr = Ndarray(ex[0], ntv_type=ex[1], shape=[2,1])
                for format in ['full', 'complete']:
                    #print(ex, format)
                    js = arr.to_json(format=format)
                    #print(js)
                    ex_rt = Ndarray.read_json(js)
                    self.assertEqual(ex_rt, arr)        

    def test_ndarray_nested2(self):    

        example =[[[[1,2], [3,4]], 'array'],
                  [[np.array([1, 2], dtype='int64'), np.array(['test1', 'test2'], dtype='str_')], 'narray'],
                  [[pd.Series([1,2,3]), pd.Series([4,5,6])], 'field'],
                  [[pd.DataFrame({'::date': pd.Series([date(1964,1,1), date(1985,2,5)]), 
                                  'names': ['john', 'eric']}),
                    pd.DataFrame({'::date': pd.Series([date(1984,1,1), date(1995,2,5)]), 
                                  'names': ['anna', 'erich']})], 'tab' ]
                  ]
        for ex in example:
            arr = Ndarray(ex[0], shape=[2], ntv_type=ex[1])
            for format in ['full', 'complete']:
                js = arr.to_json(format=format)
                #print(js)
                ex_rt = Ndarray.read_json(js)
                self.assertEqual(ex_rt, arr)        
                #print(nd_equals(ex_rt, arr),  ex_rt, ex_rt.dtype)
        
    def test_ndarray_ntvtype2(self):    

        example = [['int64[kg]', [[1, 2], [3,4]]],
                   ['int', [[1, 2], [3,4]]],
                   ['json', [1, 'two']],
                   ['month', [1, 2]],
                   ['base16', ['1F23', '236A5E']],
                   ['duration', ['P3Y6M4DT12H30M5S', 'P3Y6M4DT12H30M']],
                   ['uri', ['geo:13.4125,103.86673', 'geo:13.41,103.86']],
                   ['email', ['John Doe <jdoe@mac.example>', 'Anna Doe <adoe@mac.example>']],
                   #['$org.propertyID', ['NO2', 'NH3']]
                   ['ipv4', ['192.168.1.1', '192.168.2.5']],
                   [None, ['a', 's']],
                   #[None, 'uri'],
                   ['float', 'uri'],
                  ]
        for ex in example:
            arr = Ndarray(ex[1], ntv_type=ex[0])
            for format in ['full', 'complete']:
                js = arr.to_json(format=format)
                #print(js)
                ex_rt = Ndarray.read_json(js)
                #print(ex_rt)
                self.assertEqual(ex_rt, arr)        

    def test_ndarray_uri2(self):    
        file = 'https://raw.githubusercontent.com/loco-philippe/ntv-numpy/master/example/ex_ndarray.ntv'
        jsn = requests.get(file, allow_redirects=True).content.decode()
        #print(type(jsn), jsn)
        nda = Ndarray.read_json(jsn)
        #print(nda)
        self.assertEqual(nda, Ndarray.read_json({':ndarray': ['int64[kg]', [2, 2], [1, 2, 3, 4]]}))
        example = [['uri', 'int32', None],
                   ['uri', None, None],
                   ['uri', 'int32', [2,2]],
                   ['uri', None, [2,2]],
                  ]
        for ex in example:
            nda = Ndarray(ex[0], ex[1], ex[2])
            self.assertEqual(Ndarray.read_json(nda.to_json()), nda )
        
class Test_Xndarray(unittest.TestCase):    

    def test_xndarray_simple(self):    
        
        example =[
                  {'y': [['string', ['y1', 'y2']]]},
                  
                ]
        for ex in example:
            self.assertEqual(ex, Xndarray.read_json(ex).to_json(header=False))         
            
        example =[
                  {':xndarray': [['int64[kg]', [10, 20]]]},
                  {':xndarray': [['month', [1, 2]]]},
                  {':xndarray': [['ipv4', ['192.168.1.1', '192.168.2.5']]]},
                  {':xndarray': [['json', [1, 'two', {'three': 3}]]]},
                  {':xndarray': [['base16', [b'1F23', b'236A5E']]]},
                  {':xndarray': [['uri', ['geo:13.4125,103.86673', 'geo:13.41,103.86']]]},
                  {':xndarray': [['object', 'https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv']]}   
                ]
        for ex in example:
            #print(ex)
            self.assertEqual(ex, Xndarray.read_json(ex).to_json())    
            xnd = Xndarray.read_json(ex)
            self.assertEqual(xnd, Xndarray.read_json(xnd.to_json()))    
            
    def test_xndarray_dataset(self):    
        
        example =[[{'var1': [['object', 'https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], 
                                    ['x', 'y']]}, 'relative', 'variable'],
                  [{'var1': [['https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], 
                                    ['x', 'y']]}, 'relative', 'variable'],
                  [{'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']]}, 'absolute', 'variable'],

                  [{'ranking': [['int', [2, 2], [1, 2, 3, 4]], ['var1']]}, 'absolute', 'variable'],
                  [{'x': [['string', ['x1', 'x2']], {'test': 21}]}, 'absolute', 'namedarray'],
                  [{'y': [['string', ['y1', 'y2']]]}, 'absolute', 'namedarray'],
                  [{'z': [['string', ['z1', 'z2']], ['x']]}, 'absolute', 'variable'],
                  [{'x.mask': [['boolean', [True, False]]]}, 'absolute', 'additional'],
                  [{'x.variance': [['float64', [0.1, 0.2]]]}, 'absolute', 'additional'],
                  [{'z.variance': [['float64', [0.1, 0.2]]]}, 'absolute', 'additional'],
                  [{'unit': 'kg'}, 'undefined', 'metadata'],
                  [{'info': {'example': 'everything'}}, 'undefined', 'metadata'],
                ]
        
        for ex, mode, xtype in example:
            #print(ex)
            self.assertEqual(ex, Xndarray.read_json(ex).to_json(header=False)) 
            self.assertEqual(mode, Xndarray.read_json(ex).mode) 
            self.assertEqual(xtype, Xndarray.read_json(ex).xtype) 
            xa = Xndarray.read_json(ex)
            for format in ['full', 'complete']:
                #print(xa.to_json(format=format))
                #print(Xndarray.read_json(xa.to_json(format=format)))
                self.assertEqual(xa, Xndarray.read_json(xa.to_json(format=format)))      

        example2 =[{'var1': [['object', 'https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], 
                            ['x', 'y']]},
                   {'var1': [['https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], 
                            ['x', 'y']]},
                   {'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']]},
                   {'ranking': [[[2, 2], [1, 2, 3, 4]], ['var1']]},
                   {'x': [[['x1', 'x2']], {'test': 21}]},
                   {'y': [[['y1', 'y2']]]},
                   {'z': [[['z1', 'z2']], ['x']]},
                   {'x.mask': [[[True, False]]]},
                   {'x.variance': [[[0.1, 0.2]]]},
                   {'z.variance': [[[0.1, 0.2]]]},
                  ]
        for (ex, mode, xtype), ex2 in zip(example, example2):
            #print(ex, ex2)
            self.assertEqual(Xndarray.read_json(ex2).to_json(header=False), ex)

class Test_Xdataset(unittest.TestCase):    

    def test_xdataset_full(self):    
        
        example = {'test': {
            'var1': [['https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], 
                     ['x', 'y']],
            'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
            'ranking': [[[2, 2], [1, 2, 3, 4]], ['var1']],
            'x': [[['x1', 'x2']], {'test': 21}],
            'y': [[['y1', 'y2']]],
            'z': [[['z1', 'z2']], ['x']],
            'x.mask1': [[[True, False]]],
            'x.variance': [[[0.1, 0.2]]],
            'z.variance': [[[0.1, 0.2]]],
            'unit': 'kg',
            'info': {'example': 'everything'}}}
        
        notype = [True, False, True, True, True, True, True, True, True, True, True]
        xds = Xdataset.read_json(example)        
        self.assertEqual(xds.to_json(notype=notype, noshape=True, header=False), example)                                          
        self.assertEqual(xds.dimensions, ('x', 'y'))
        self.assertEqual(xds.partition, {'coordinates': ['ranking', 'z'],
         'data_vars': ['var1', 'var2'], 'metadata': ['info', 'unit'],
         'additionals': ['x.mask1', 'x.variance', 'z.variance'], 'dimensions': ['x', 'y']})
        
        xdim = Xdataset(xds[xds.dimensions])
        self.assertEqual(xdim.to_json(novalue=True, noshape=True), {':xdataset': {
                         'x': [['string', ['-']], {'test': 21}],
                         'y': [['string', ['-']]]}})        
                                  
    def test_xdataset_info(self):    
        
        xd = Xdataset([Xndarray('example', np.array(['x1', 'x2']))], 'test')
        self.assertEqual(xd.info, {'name': 'test', 'xtype': 'group', 'validity': 'valid',
                                   'data_arrays': ['example'], 'width': 1})

        example = {'test': {
            'var1': [['https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], ['x', 'y']],
            'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
            'ranking': [[[2, 2], [1, 2, 3, 4]], ['var2']],
            'x': [[['x1', 'x2']], {'test': 21}],
            'y': [[['y1', 'y2']]],
            'z': [[['z1', 'z2']], ['x']],
            'z_bis': [[['z1_bis', 'z2_bis']]],
            'x.mask1': [[[True, False]], ['x']],
            'x.variance': [[[0.1, 0.2]], ['x']],
            'z.variance': [[[0.1, 0.2]], ['x']],
            'unit': 'kg',
            'info': {'example': 'everything'}}}
        xd = Xdataset.read_json(example)
        self.assertEqual(xd.info, { 'name': 'test', 'xtype': 'group',
                                    'data_vars': ['var1', 'var2'],
                                    'data_arrays': ['z_bis'],
                                    'dimensions': ['x', 'y'],
                                    'coordinates': ['ranking', 'z'],
                                    'additionals': ['x.mask1', 'x.variance', 'z.variance'],
                                    'metadata': ['info', 'unit'],
                                    'validity': 'undefined', 'width': 12})

        del(xd[('var1', 'z_bis')])
        self.assertEqual(xd.info, { 'name': 'test', 'xtype': 'mono',
                                    'data_vars': ['var2'],
                                    'dimensions': ['x', 'y'],
                                    'length' : 4,
                                    'coordinates': ['ranking', 'z'],
                                    'additionals': ['x.mask1', 'x.variance', 'z.variance'],
                                    'metadata': ['info', 'unit'],
                                    'validity': 'valid', 'width': 10})        

        example = {'test': {
            'var1': [['float[m3]', [2, 2], 'path/var1.ntv'], ['x', 'y']],
            'var2': [['float[kg]', [2, 2], 'path/var2.ntv'], ['x', 'y']],
            'ranking': [[[2, 2], 'path/ranking.ntv'], ['var2']],
            'x': [['path/x.ntv'], {'test': 21}],
            'y': [['path/y.ntv']],
            'z': [['path/z.ntv'], ['x']],
            'z_bis': [['path/z_bis.ntv']],
            'x.mask1': [['path/x.mask1.ntv'], ['x']],
            'x.variance': [['path/x.variance.ntv'], ['x']],
            'z.variance': [['path/z.variance.ntv'], ['x']],
            'info': {'path': 'https://github.com/loco-philippe/ntv-numpy/tree/main/example/'}}}
       
        xd = Xdataset.read_json(example)
        self.assertEqual(xd.info, { 'name': 'test', 'xtype': 'group',
                                    'data_vars': ['var1', 'var2'],
                                    'data_arrays': ['z_bis'],
                                    'dimensions': ['x', 'y'],
                                    'coordinates': ['ranking', 'z'],
                                    'additionals': ['x.mask1', 'x.variance', 'z.variance'],
                                    'metadata': ['info'],
                                    'validity': 'undefined', 'width': 11})

    def test_xdataset_DataArray(self):    
        
        examples = [
            {'test': {
                'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
                'unit': 'kg', 'info': {'example': 'everything'},
                'ranking': [[[2, 2], [1, 2, 3, 4]], ['var2']],  #!!!
                'x': [[['x1', 'x2']], {'test': 21}],
                'y': [['date', ['2021-01-01', '2022-02-02']]],
                'z': [[['z1', 'z2']], ['x']],
                #'z_bis': [[['z1_bis', 'z2_bis']]],
                'x.mask1': [[[True, False]]],
                'x.variance': [[[0.1, 0.2]]],
                'z.variance': [[[0.1, 0.2]]]
            }},
            {'test': {
                'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
                'var2.variance': [[[2, 2], [0.1, 0.2, 0.3, 0.4]]],
                'var2.mask1': [[[True, False]], ['x']],
                'var2.mask2': [[[2, 2], [True, False, False, True]]],

                'ranking': [['month', [2, 2], [1, 2, 3, 4]], ['var2']], #!!!
                

                'x': [['string', ['23F0AE', '578B98']], {'test': 21}],
                'x.mask1': [[[True, False]]],

                'y': [['date', ['2021-01-01', '2022-02-02']]],

                'z': [['float', [10, 20]], ['x']],
                #'z_bis': [[['z1_bis', 'z2_bis']]],
                'z.variance': [[[0.1, 0.2]]],

                'unit': 'kg',
                'info': {'example': 'everything'}
            }}]
        for example in examples:
            xd = Xdataset.read_json(example) 
            xd2 = Xdataset.from_xarray(xd.to_xarray(dataset=False))
            self.assertEqual(xd, xd2)
            xd2 = Xdataset.from_xarray(xd.to_xarray())
            self.assertEqual(xd, xd2)        

        examples = [xr.DataArray(np.array([1,2,3,4]).reshape([2,2])),
                    xr.Dataset({'var': (['date', 'y'], np.array([1,2,3,4]).reshape([2,2]))},
                               coords={'date': np.array(['2021-02-04','2022-02-04'], dtype='datetime64[ns]'),
                                       'y': np.array([10, 20])}),
                    xr.Dataset({'var': (['date', 'y'], np.array([1,2,3,4]).reshape([2,2]))},
                               coords={'date': (['date'], np.array(['2021-02-04','2022-02-04'], dtype='datetime64[ns]'), {'ntv_type': 'date'}),
                                       'y': np.array([10, 20])}),
                    xr.Dataset({'var': (['date', 'y'], np.array([1,2,3,4]).reshape([2,2]))},
                               coords={'date': (['date'], np.array(['2021-02-04','2022-02-04'], dtype='datetime64[ns]'), {'ntv_type': 'date'}),
                                       'y': np.array([Point([1,2]), Point([3,4])])})]

        for xar in examples:
            xd = Xdataset.from_xarray(xar)
            #print(xd.to_json())
            for dts in [False, True]:
                xar2 = xd.to_xarray(dataset=dts)
                xd2 = Xdataset.from_xarray(xar2)
                #print(xd2.to_json())
                self.assertEqual(xd, xd2)
                self.assertEqual(xd.to_json(), xd2.to_json())

    def test_xdataset_scipp(self):    
        
        examples = [{ #'test': {
            'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
            'var2.variance': [[[2, 2], [0.1, 0.2, 0.3, 0.4]]],
            'var2.mask1': [[[True, False]], ['x']],
            'var2.mask2': [[[2, 2], [True, False, False, True]]],

            'ranking': [['month', [2, 2], [1, 2, 3, 4]], ['var2']],
            

            'x': [['string', ['23F0AE', '578B98']]], #, {'test': 21}],
            #'x.mask1': [[[True, False]]],

            'y': [['date', ['2021-01-01', '2022-02-02']]],

            'z': [['float', [10, 20]], ['x']],
            #'z_bis': [[['z1_bis', 'z2_bis']]],
            'z.variance': [['float', [0.1, 0.2]]],

            #'unit': 'kg',
            #'info': {'example': 'everything'}
            } #}
            ]
        for example in examples:
            xd = Xdataset.read_json(example) 
            xd2 = Xdataset.from_scipp(xd.to_scipp(dataset=False, datagroup=False))
            self.assertEqual(xd, xd2)
            xd2 = Xdataset.from_scipp(xd.to_scipp(dataset=False))
            self.assertEqual(xd, xd2)
            xd2 = Xdataset.from_scipp(xd.to_scipp())
            self.assertEqual(xd, xd2)        

    def test_xdataset_mixte(self):    
        
        examples = [{ 'test:xdataset': {
            'var1': [['https://github.com/loco-philippe/ntv-numpy/tree/main/example/ex_ndarray.ntv'], ['x', 'y']],    
            'var2': [['float[kg]', [2, 2], [10.1, 0.4, 3.4, 8.2]], ['x', 'y']],
            'var2.variance': [[[2, 2], [0.1, 0.2, 0.3, 0.4]]],
            'var2.mask1': [[[True, False]], ['x']],
            'var2.mask2': [[[2, 2], [True, False, False, True]]],

            'ranking': [['month', [2, 2], [1, 2, 3, 4]], ['var2']], #!!!


            'x': [['string', ['23F0AE', '578B98']]], #, {'test': 21}],
            'x.mask1': [[[True, False]]],

            'y': [['date', ['2021-01-01', '2022-02-02']]],

            'z': [['float', [10, 20]], ['x']],
            'z_bis': [[['z1_bis', 'z2_bis']]],
            'z.uncertainty': [[[0.1, 0.2]]],
            'z.variance': [['float', [0.1, 0.2]]],

            'info': {'example': 'everything'}
            } }
            ]
        for example in examples:
            xd = Xdataset.read_json(example) 
            xd_sc = Xdataset.from_scipp(xd.to_scipp(dataset=False))
            xd_xr = Xdataset.from_xarray(xd.to_xarray(dataset=False))
            self.assertTrue(xd == xd_sc == xd_xr)
            xd_sc = Xdataset.from_scipp(xd.to_scipp())
            xd_xr = Xdataset.from_xarray(xd.to_xarray())
            self.assertTrue(xd == xd_sc == xd_xr)

if __name__ == '__main__':    
    unittest.main(verbosity=2)
                                    